

from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting
import random
import sys


class ULRandomInt(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.max_value = None
        self.min_value = None
        self.OUT_A = ULOutSocket(self, self._get_output)

    def _get_output(self):
        socket = self.get_output('result')
        if socket is None:
            min_value = self.get_input(self.min_value)
            max_value = self.get_input(self.max_value)
            if is_waiting(max_value, min_value):
                return STATUS_WAITING
            if min_value > max_value:
                min_value, max_value = max_value, min_value
            if min_value == max_value:
                min_value = -sys.maxsize
                max_value = sys.maxsize

            return self.set_output(
                'result',
                random.randint(min_value, max_value)
            )
        return socket

    def evaluate(self):
        self._set_ready()



from mathutils import Vector
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING, is_invalid


class ULVectorAbsolute(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.input_v = None
        self.output_v = Vector()
        self.OUTV = ULOutSocket(self, self.get_out_v)

    def get_out_v(self):
        vec = self.get_input(self.input_v)
        if is_invalid(vec):
            return STATUS_WAITING
        vec = vec.copy()
        vec.x = abs(vec.x)
        vec.y = abs(vec.y)
        vec.z = abs(vec.z)
        return vec

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULGetLightColor(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.lamp = None
        self.color = 0
        self.COLOR = ULOutSocket(self, self.get_color)

    def get_color(self):
        lamp = self.get_input(self.lamp)
        if is_waiting(lamp):
            return STATUS_WAITING
        light = lamp.blenderObject.data
        return light.color

    def evaluate(self):
        self._set_ready()



from uplogic.nodes import ULActionNode
from uplogic.utils import is_waiting
from uplogic.utils import not_met
from uplogic.utils import move_to


# Action "Move To": an object will follow a point
class ULMoveTo(ULActionNode):

    def __init__(self):
        ULActionNode.__init__(self)
        # list of parameters of this action
        self.condition = None
        self.moving_object = None
        self.destination_point = None
        self.speed = None
        self.dynamic = None
        self.distance = None

    def evaluate(self):  # the actual execution of this cell
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        moving_object = self.get_input(self.moving_object)
        destination_point = self.get_input(self.destination_point)
        speed = self.get_input(self.speed)
        distance = self.get_input(self.distance)
        dynamic = self.get_input(self.dynamic)
        if hasattr(destination_point, 'worldPosition'):
            destination_point = destination_point.worldPosition
        if is_waiting(
            moving_object,
            destination_point,
            speed,
            distance,
            dynamic
        ):
            return
        self._set_ready()
        self._set_value(move_to(
            moving_object,
            destination_point,
            speed,
            self.network.time_per_frame,
            dynamic,
            distance
        ))



from bge import logic
from uplogic.nodes import ULConditionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULGamepadButtonUp(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.pulse = False
        self.button = 0
        self.index = None
        self._last_value = False
        self._up_value = None
        self.BUTTON = ULOutSocket(self, self.get_button)
        self.initialized = False

    def get_button(self):
        socket = self.get_output('button')
        if socket is None:
            pressed = False
            index = self.get_input(self.index)
            if logic.joysticks[index]:
                joystick = logic.joysticks[index]
            else:
                return STATUS_WAITING
            if is_invalid(joystick):
                return STATUS_WAITING

            button_down = (
                True if
                self.button in joystick.activeButtons else
                False
            )

            if button_down != self._last_value and not button_down:
                if not self.pulse and not self.initialized:
                    self.initialized = True
                pressed = True
            elif (pressed and self.initialized) or button_down:

                pressed = False
                self.initialized = False
            elif not (self.initialized and button_down) and pressed:
                pressed = True

            self._last_value = button_down
            return self.set_output(
                'button',
                pressed
            )
        return socket

    def evaluate(self):
        self._set_ready()



from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULTranslate(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.moving_object = None
        self.local = None
        self.vect = None
        self.speed = None
        self._t = None
        self._old_values = None

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            self._set_value(False)
            return self._set_ready()
        moving_object = self.get_input(self.moving_object)
        vect = self.get_input(self.vect)
        dx = vect.x
        dy = vect.y
        dz = vect.z
        speed = self.get_input(self.speed)
        local = self.get_input(self.local)
        if is_waiting(
            vect,
            dx,
            dy,
            dz,
            speed,
            local
        ):
            return
        if is_invalid(moving_object):
            return
        self._set_ready()
        if dx is None:
            return
        if dy is None:
            return
        if dz is None:
            return
        if speed is None:
            return
        if local is None:
            return
        new_values = (moving_object, dx, dy, dz, speed, local)
        if new_values != self._old_values:
            start_pos = (
                moving_object.localPosition if
                local else moving_object.worldPosition
            )
            end_pos = Vector((
                start_pos.x + dx, start_pos.y + dy, start_pos.z + dz
            ))
            distance = (start_pos - end_pos).length
            time_to_destination = distance / speed
            t_speed = 1.0 / time_to_destination
            self._old_values = new_values
            self._start_pos = start_pos.copy()
            self._end_pos = end_pos.copy()
            self._d_pos = (end_pos - start_pos)
            self._t_speed = t_speed
            self._t = 0.0
        t = self._t
        dt = self._t_speed * self.network.time_per_frame
        t += dt
        if t >= 1.0:
            self._set_value(True)
            self._t = 0.0
            if local:
                moving_object.localPosition = self._end_pos.copy()
            else:
                moving_object.worldPosition = self._end_pos.copy()
        else:
            npos = self._start_pos + (self._d_pos * t)
            if local:
                moving_object.localPosition = npos
            else:
                moving_object.worldPosition = npos
            self._t = t
            self._set_value(False)



'''TODO: Documentation
'''

from os.path import isfile
from bge import logic
from bge.types import KX_GameObject as GameObject
from mathutils import Vector
from uplogic.audio import ULAudioSystem
from uplogic.audio import get_audio_system
from uplogic.events import schedule_callback
from uplogic.utils import debug, interpolate
import aud


class ULReverb():
    """Reverb sound added by `ULSound3D` on demand.
    """

    volume: float

    def __init__(
        self,
        parent,
        sound,
        handle
    ):
        self.volume = 0
        self.parent = parent
        self.handle = handle
        self.aud_system = parent.aud_system
        self.samples = []
        schedule_callback(self.add_sample, 1/60, sound)

    def add_sample(self, sound):
        handle = self.handle
        sample = self.aud_system.device.play(sound)
        self.samples.append(sample)
        sample.loop_count = handle.loop_count
        sample.position = handle.position - (.0001 * len(self.samples))
        sample.relative = handle.relative
        sample.location = handle.location
        sample.velocity = handle.velocity
        sample.attenuation = handle.attenuation
        ori = self.parent.speaker.worldOrientation.to_quaternion()
        ori.negate()
        sample.orientation = ori
        sample.pitch = handle.pitch
        sample.volume = 0
        sample.distance_reference = handle.distance_reference
        sample.distance_maximum = handle.distance_maximum
        sample.cone_angle_inner = handle.cone_angle_inner
        sample.cone_angle_outer = handle.cone_angle_outer
        sample.cone_volume_outer = handle.cone_volume_outer
        if len(self.samples) < 30:
            schedule_callback(self.add_sample, 1/60, sound)

    def update(self):
        sample_count = self.aud_system.bounces
        use_reverb = (
            self.aud_system.reverb
        )
        handle = self.handle
        if not use_reverb or sample_count == 0:
            if self.volume < .001:
                return
            else:
                self.volume = interpolate(self.volume, 0, .1)
        else:
            parent = self.parent
            target_vol = (
                parent.volume / 10 if
                parent.occluded else
                parent.volume / 2
            )
            self.volume = interpolate(self.volume, target_vol, .1)
        for idx, sample in enumerate(self.samples):
            if not sample.status:
                sample.stop()
                continue
            if idx > sample_count:
                sample.volume = 0
                continue
            mult = idx/sample_count
            loc = handle.location
            lloc = self.aud_system.device.listener_location
            loc = (loc[0]-lloc[0], loc[1]-lloc[1], loc[2]-lloc[2])
            sample.location = (
                -(loc[0]-lloc[0]),
                -(loc[1]-lloc[1]),
                -(loc[2]-lloc[2])
            )
            sample.velocity = handle.velocity
            sample.attenuation = handle.attenuation
            ori = self.parent.speaker.worldOrientation.to_quaternion()
            ori.negate()
            sample.orientation = ori
            sample.distance_maximum = handle.distance_maximum
            sample.cone_angle_inner = handle.cone_angle_inner
            sample.pitch = handle.pitch
            sample.volume = (1-(handle.volume * (mult**2)))*.5 * self.volume * self.aud_system.volume
            sample.cone_volume_outer = handle.cone_volume_outer


class ULSound():
    """Base class for 2D and 3D Sounds
    """

    sound = None
    """Internal `aud.Sound` instance."""
    finished: bool
    """Whether this sound has finished playing."""
    pitch: float
    """Pitch (Frequency Shift)."""
    volume: float
    aud_system: ULAudioSystem

    @property
    def position(self):
        if self.sound:
            return self.sound.position
    
    @position.setter
    def position(self, val):
        if self.sound:
            self.sound.position = val

    def stop(self):
        '''TODO: Documentation
        '''
        self.sound.stop()

    def pause(self):
        self.sound.pause()

    def resume(self):
        self.sound.resume()


class ULSound2D(ULSound):
    '''Non-spacial sound, e.g. Music or Voice-Overs.\n
    This class allows for modification of pitch and volume while playing.

    :param `file`: Path to the sound file.
    :param `volume`: Initial volume.
    :param `pitch`: Initial pitch.
    :param `loop_count`: Plays the sound this many times (0 for once, -1 for endless).
    :param `aud_sys`: Audiosystem to play this sound on.
    '''

    sound: aud.Handle

    def __init__(
        self,
        file: str = '',
        volume: float = 1,
        pitch: float = 1,
        loop_count: int = 0,
        lowpass=False,
        aud_sys: str = 'default'
    ):
        self.file = file
        self._volume = 1
        self.finished = False
        if not (file):
            return
        self.aud_system = get_audio_system(aud_sys)
        soundfile = logic.expandPath(file)
        if not isfile(soundfile):
            debug(f'Soundfile {soundfile} could not be loaded!')
            return
        sound = self.soundfile = aud.Sound(soundfile)
        lowpass = self.aud_system.lowpass or lowpass
        if lowpass:
            sound = self.soundfile = sound.lowpass(lowpass, .5)
        device = self.aud_system.device
        self.sound = handle = device.play(sound)

        handle.relative = True
        handle.loop_count = loop_count
        self.aud_system.add(self)
        self.volume = volume
        self.pitch = pitch
        self._lowpass = False
        self.lowpass = self.aud_system.lowpass

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, val):
        if self.sound and self.sound.status:
            self.sound.volume = val * self.aud_system.volume
        self._volume = val

    @property
    def pitch(self):
        if self.sound and self.sound.status:
            return self.sound.pitch / logic.getTimeScale()

    @pitch.setter
    def pitch(self, val):
        if self.sound and self.sound.status:
            self.sound.pitch = val * logic.getTimeScale()

    @property
    def lowpass(self):
        return self._lowpass

    @lowpass.setter
    def lowpass(self, val):
        if self._lowpass == val:
            return
        self._lowpass = val
        sound = self.soundfile
        if val:
            sound = sound.lowpass(val, .5)
        sound = self.aud_system.device.play(sound)
        sound.loop_count = self.sound.loop_count
        sound.position = self.sound.position
        sound.volume = self.sound.volume
        sound.pitch = self.sound.pitch
        schedule_callback(self.sound.stop, .1)
        self.sound = sound

    def update(self):
        '''TODO: Documentation
        '''
        handle = self.sound
        if not handle.status:
            self.finished = True
            self.aud_system.remove(self)
            return


class ULSound3D(ULSound):
    '''Spacial sound, e.g. World Effects or Voices.\n
    This class allows for modification of pitch and volume as well as other attributes while playing.
    '''
    sounds: list
    speaker: GameObject
    occlusion: bool
    location: Vector
    cone_outer_volume: float
    transition: float
    soundpath: str
    reverb: bool
    bounces: int
    _clear_sound: float
    _sustained: float
    reverb_samples: ULReverb

    def __init__(
        self,
        speaker: GameObject = None,
        file: str = '',
        occlusion: bool = False,
        transition_speed: float = .1,
        cutoff_frequency: float = .1,
        loop_count: int = 0,
        pitch: float = 1,
        volume: float = 1,
        reverb=False,
        attenuation: float = 1,
        distance_ref: float = 1,
        cone_angle: list[float] = [360, 360],
        cone_outer_volume: float = 0,
        aud_sys: str = 'default'
    ):
        self.file = file
        self.finished = False
        if not (file and speaker):
            return
        self._clear_sound = 1
        self._sustained = 1
        self.occluded = False
        self.sounds = []
        self.reverb_samples = None
        self.aud_system = get_audio_system(aud_sys)
        self.speaker = speaker
        self.reverb = reverb
        self.occlusion = occlusion
        self.volume = volume
        self.pitch = pitch
        self.cone_outer_volume = cone_outer_volume
        master_volume = self.aud_system.volume
        self.transition = transition_speed
        soundfile = logic.expandPath(file)
        if not isfile(soundfile):
            debug(f'Soundfile {soundfile} could not be loaded!')
            return
        sound = self.soundpath = aud.Sound(soundfile)
        device = self.aud_system.device
        handle = device.play(sound)
        if occlusion:
            soundlow = aud.Sound.lowpass(sound, 4400 * cutoff_frequency, .5)
            handlelow = device.play(soundlow)
            self.handles = [speaker, [handle, handlelow]]
        else:
            self.handles = [speaker, [handle]]
        for handle in self.handles[1]:
            handle.relative = False
            handle.location = speaker.worldPosition
            if speaker.mass:
                handle.velocity = getattr(
                    speaker,
                    'worldLinearVelocity',
                    Vector((0, 0, 0))
                )
            handle.attenuation = attenuation
            handle.orientation = speaker.worldOrientation.to_quaternion()
            handle.pitch = pitch
            handle.volume = volume * master_volume
            handle.distance_reference = distance_ref
            handle.distance_maximum = 1000
            handle.cone_angle_inner = cone_angle[0]
            handle.cone_angle_outer = cone_angle[1]
            handle.loop_count = loop_count
            handle.cone_volume_outer = cone_outer_volume * volume * master_volume
        if self.reverb:
            self.reverb_samples = ULReverb(
                self,
                sound,
                self.handles[1][0]
            )
        self.aud_system.add(self)

    def update(self):
        '''TODO: Documentation
        '''
        aud_system = self.aud_system
        speaker = self.speaker
        if not speaker or speaker.invalid:
            self.finished = True
            aud_system.remove(self)
            return
        location = speaker.worldPosition
        for i, handle in enumerate(self.handles[1]):
            if not handle.status:
                self.finished = True
                aud_system.remove(self)
                return
            handle.pitch = self.pitch * logic.getTimeScale()
            handle.location = location
            handle.orientation = (
                speaker
                .worldOrientation
                .to_quaternion()
            )
            if 'volume' in dir(self.speaker.blenderObject.data):
                handle.velocity = Vector((0, 0, 0))
            else:
                handle.velocity = getattr(speaker, 'worldLinearVelocity', Vector((0, 0, 0)))
            if self.occlusion:
                transition = self.transition
                cam = self.aud_system.listener
                occluder, point, normal = cam.rayCast(
                    location,
                    cam.worldPosition,
                    speaker.getDistanceTo(cam.worldPosition),
                    xray=False
                )
                occluded = self.occluded = False
                penetration = 1
                while occluder:
                    if occluder is speaker:
                        break
                    sound_occluder = occluder.blenderObject.get(
                        'sound_occluder',
                        True
                    )
                    if sound_occluder:
                        occluded = self.occluded = True
                        block = occluder.blenderObject.get(
                            'sound_blocking',
                            .1
                        )
                        if penetration > 0:
                            penetration -= block
                        else:
                            penetration = 0
                    occluder, point, normal = occluder.rayCast(
                        location,
                        point,
                        speaker.getDistanceTo(point),
                        xray=False
                    )
                cs = self._clear_sound
                if occluded and cs > 0:
                    self._clear_sound -= transition
                elif not occluded and cs < 1:
                    self._clear_sound += transition
                if self._clear_sound < 0:
                    self._clear_sound = 0
                sustained = self._sustained
                if sustained > penetration:
                    self._sustained -= transition / 10
                elif sustained < penetration:
                    self._sustained += transition / 10
                mult = (
                    cs * sustained
                    if not i
                    else (1 - cs) * sustained
                )
                master_volume = self.aud_system.volume
                handle.volume = self.volume * mult * master_volume
                handle.cone_volume_outer = (
                    self.cone_outer_volume *
                    self.volume *
                    mult *
                    master_volume
                )
        if self.reverb_samples:
            self.reverb_samples.update()

    def stop(self):
        '''TODO: Documentation
        '''
        for sound in self.sounds:
            sound.stop()



from bge.types import KX_GameObject as GameObject
from bpy.types import Material
from bpy.types import Node
from bpy.types import NodeSocket
from bpy.types import NodeGroup
import bpy


def get_material_socket(
    material: str or Material,
    node: Node,
    socket: NodeSocket,
    game_object: GameObject = None
):
    if isinstance(material, str):
        material = bpy.data.materials[material]
    if isinstance(material, int):
        material = game_object.blenderObject.material_slots[0].material
    return (
        material
        .node_tree
        .nodes[node]
        .inputs[socket]
        .default_value
    )


def set_material_socket(
    material: str or Material,
    node: Node,
    socket: NodeSocket,
    value,
    game_object: GameObject = None
):
    if isinstance(material, str):
        material = bpy.data.materials[material]
    if isinstance(material, int):
        material = game_object.blenderObject.material_slots[0].material
    (
        material
        .node_tree
        .nodes[node]
        .inputs[socket]
        .default_value
    ) = value


def modify_material_socket(
    material: str or Material,
    node: Node,
    socket: NodeSocket,
    value,
    game_object: GameObject = None
):
    if isinstance(material, str):
        material = bpy.data.materials[material]
    if isinstance(material, int):
        material = game_object.blenderObject.material_slots[0].material
    (
        material
        .node_tree
        .nodes[node]
        .inputs[socket]
        .default_value
    ) += value


def set_world_socket(world, node, socket, value, output=False):
    if isinstance(world, str):
        world = bpy.data.worlds[world]
    if output:
        (
            world
            .node_tree
            .nodes[node]
            .outputs[socket]
            .default_value
        ) = value
    else:
        (
            world
            .node_tree
            .nodes[node]
            .inputs[socket]
            .default_value
        ) = value


def modify_world_socket(world, node, socket, value, output=False):
    if isinstance(world, str):
        world = bpy.data.worlds[world]
    if output:
        (
            world
            .node_tree
            .nodes[node]
            .outputs[socket]
            .default_value
        ) = value
    else:
        (
            world
            .node_tree
            .nodes[node]
            .inputs[socket]
            .default_value
        ) += value


def get_world_socket(world, node, socket, output=False):
    if isinstance(world, str):
        world = bpy.data.worlds[world]
    return (
        world
        .node_tree
        .nodes[node]
        .outputs[socket]
        .default_value
    ) if output else (
        world
        .node_tree
        .nodes[node]
        .inputs[socket]
        .default_value
    )


def get_geom_socket(
    tree: str or NodeGroup,
    node: str or Node,
    socket: int or NodeSocket
):
    if isinstance(tree, str):
        tree = bpy.data.node_groups[tree]
    return (
        tree
        .nodes[node]
        .inputs[socket]
        .default_value
    )


def set_geom_socket(
    tree: str or NodeGroup,
    node: str or Node,
    socket: int or NodeSocket,
    value
):
    if isinstance(tree, str):
        tree = bpy.data.node_groups[tree]
    (
        tree
        .nodes[node]
        .inputs[socket]
        .default_value
    ) = value


def modify_geom_socket(
    tree: str or NodeGroup,
    node: str or Node,
    socket: int or NodeSocket,
    value
):
    if isinstance(tree, str):
        tree = bpy.data.node_groups[tree]
    (
        tree
        .nodes[node]
        .inputs[socket]
        .default_value
    ) += value


from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULWithinRange(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.range = None
        self.operator = None
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        v = self.get_input(self.value)
        r = self.get_input(self.range)
        if is_waiting(v, r):
            return STATUS_WAITING
        value = self.calc_range(self.operator, v, r)
        if (v is None) or (r is None):
            return STATUS_WAITING
        else:
            return value

    def calc_range(self, op, v, r):
        if op == 'OUTSIDE':
            return True if (v < r.x or v > r.y) else False
        if op == 'INSIDE':
            return True if (r.x < v < r.y) else False

    def evaluate(self):
        self._set_ready()



from bge import constraints
from bge import logic
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import is_invalid
import bpy


class ULCharacterInfo(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.max_jumps = None
        self.cur_jump = None
        self.gravity = None
        self.walk_dir = None
        self.on_ground = None
        self.local = False
        self.MAX_JUMPS = ULOutSocket(self, self.get_max_jumps)
        self.CUR_JUMP = ULOutSocket(self, self.get_current_jump)
        self.GRAVITY = ULOutSocket(self, self.get_gravity)
        self.WALKDIR = ULOutSocket(self, self.get_walk_dir)
        self.ON_GROUND = ULOutSocket(self, self.get_on_ground)

    def get_max_jumps(self):
        return self.physics.maxJumps

    def get_current_jump(self):
        return self.physics.jumpCount

    def get_gravity(self):
        return self.physics.gravity

    def get_walk_dir(self):
        socket = self.get_output('walk_dir')
        if socket is None:
            physics = self.physics
            wdir = (
                physics.walkDirection @ self.owner.worldOrientation
                if self.local else
                physics.walkDirection
            )
            return self.set_output(
                'walk_dir',
                wdir * bpy.data.scenes[
                    logic.getCurrentScene().name
                ].game_settings.physics_step_sub
            )
        return socket

    def get_on_ground(self):
        return self.physics.onGround

    def evaluate(self):
        game_object = self.owner = self.get_input(self.game_object)
        if is_invalid(game_object):
            return
        self.physics = constraints.getCharacter(game_object)
        self._set_ready()



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULChildByName(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.from_parent = None
        self.child = None
        self.CHILD = ULOutSocket(self, self.get_child)

    def get_child(self):
        socket = self.get_output('child')
        if socket is None:
            parent = self.get_input(self.from_parent)
            child_name = self.get_input(self.child)
            if is_invalid(parent, child_name):
                return STATUS_WAITING
            return self.set_output(
                'child',
                parent.childrenRecursive.get(child_name)
            )
        return socket

    def evaluate(self):
        self._set_ready()



from mathutils import Vector
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid


class ULObjectDataVertices(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.OUT = ULOutSocket(self, self.get_data)

    def get_data(self):
        socket = self.get_output('data')
        if socket is None:
            obj = self.get_input(self.game_object)
            if is_invalid(obj):
                return STATUS_WAITING
            offset = obj.worldPosition
            return self.set_output(
                'data',
                (
                    sorted(
                        [Vector(v.co) + offset for v in (
                            obj
                            .blenderObject
                            .data
                            .vertices
                        )]
                    )
                )
            )
        return socket

    def evaluate(self):
        self._set_ready()



from bge import render
from bge import logic
from mathutils import Vector
from uplogic.nodes import ULActionNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import is_invalid
from uplogic.utils import not_met
from uplogic.utils import interpolate


class ULMouseLook(ULActionNode):
    x = None
    y = None
    screen_center = None
    center = None
    mouse = None

    def __init__(self):
        ULActionNode.__init__(self)
        self.axis = None
        self.condition = None
        self.game_object_x = None
        self.game_object_y = None
        self.inverted = None
        self.sensitivity = None
        self.use_cap_z = None
        self.cap_z = None
        self.use_cap_z = None
        self.cap_y = None
        self.smooth = None
        self.initialized = False
        self._x = 0
        self._y = 0
        self.done = None
        self.OUT = ULOutSocket(self, self.get_done)
        self.use_local_head = False
        self.get_data()
        self.mouse.position = self.screen_center

    def get_done(self):
        return self.done

    def get_x_obj(self):
        game_object_x = self.get_input(self.game_object_x)
        return game_object_x

    def get_y_obj(self):
        game_object_y = self.get_input(self.game_object_y)
        if is_invalid(game_object_y):
            game_object_y = self.get_x_obj()
        elif game_object_y is not self.get_x_obj():
            self.use_local_head = True
        return game_object_y

    def get_data(self):
        self.x = render.getWindowWidth()//2
        self.y = render.getWindowHeight()//2
        self.screen_center = (
            self.x / render.getWindowWidth(),
            self.y / render.getWindowHeight()
        )
        self.center = Vector(self.screen_center)
        self.mouse = logic.mouse

    def evaluate(self):
        self.done = False
        self.get_data()
        condition = self.get_input(self.condition)
        if not_met(condition):
            self.initialized = False
        elif not self.initialized:
            self.mouse.position = self.screen_center
            self.initialized = True
            return
        game_object_x = self.get_x_obj()
        game_object_y = self.get_y_obj()
        sensitivity = self.get_input(self.sensitivity) * 1000
        use_cap_z = self.get_input(self.use_cap_z)
        use_cap_y = self.get_input(self.use_cap_y)
        cap_z = self.get_input(self.cap_z)
        lowercapX = cap_z.y
        uppercapX = cap_z.x
        cap_y = self.get_input(self.cap_y)
        lowercapY = cap_y.x
        uppercapY = cap_y.y
        inverted = self.get_input(self.inverted)
        smooth = 1 - (self.get_input(self.smooth) * .99)
        self._set_ready()

        if is_invalid(game_object_x):
            return

        if condition:
            mouse_position = Vector(self.mouse.position)
            offset = (mouse_position - self.center) * -0.002
        else:
            offset = Vector((0, 0))

        if inverted.get('y', False) is False:
            offset.y = -offset.y
        if inverted.get('x', False) is True:
            offset.x = -offset.x
        offset *= sensitivity

        self._x = offset.x = interpolate(self._x, offset.x, smooth)
        self._y = offset.y = interpolate(self._y, offset.y, smooth)

        if use_cap_z:
            objectRotation = game_object_x.localOrientation.to_euler()

            if objectRotation.z + offset.x > uppercapX:
                offset.x = 0
                objectRotation.z = uppercapX
                game_object_x.localOrientation = objectRotation.to_matrix()

            if objectRotation.z + offset.x < lowercapX:
                offset.x = 0
                objectRotation.z = lowercapX
                game_object_x.localOrientation = objectRotation.to_matrix()

        game_object_x.applyRotation((0, 0, offset.x), self.use_local_head)

        rot_axis = 1 - self.axis
        if use_cap_y:
            objectRotation = game_object_y.localOrientation.to_euler()

            if objectRotation[rot_axis] + offset.y > uppercapY:
                objectRotation[rot_axis] = uppercapY
                game_object_y.localOrientation = objectRotation.to_matrix()
                offset.y = 0

            if objectRotation[rot_axis] + offset.y < lowercapY:
                objectRotation[rot_axis] = lowercapY
                game_object_y.localOrientation = objectRotation.to_matrix()
                offset.y = 0

        rot = [0, 0, 0]
        rot[1-self.axis] = offset.y
        game_object_y.applyRotation((*rot, ), True)
        if self.mouse.position != self.screen_center and condition:
            self.mouse.position = self.screen_center
        self.done = True



from .shader import ULFilter
from bge import logic


glsl = """
/*
SSAO GLSL shader v1.2
assembled by Martins Upitis (martinsh) (devlog-martinsh.blogspot.com)
original technique is made by Arkano22 (www.gamedev.net/topic/550699-ssao-no-halo-artifacts/)

changelog:
1.2 - added fog calculation to mask AO. Minor fixes.
1.1 - added spiral sampling method from here:
(http://www.cgafaq.info/wiki/Evenly_distributed_points_on_sphere)
*/
uniform sampler2D bgl_DepthTexture;
uniform sampler2D bgl_RenderedTexture;
uniform float bgl_RenderedTextureWidth;
uniform float bgl_RenderedTextureHeight;
in vec4 bgl_TexCoord;

#define PI    3.14159265

float width = bgl_RenderedTextureWidth; //texture width
float height = bgl_RenderedTextureHeight; //texture height

vec2 texCoord = bgl_TexCoord.xy;

//------------------------------------------
//general stuff

//make sure that these two values are the same for your camera, otherwise distances will be wrong.

uniform float znear; //Z-near
uniform float zfar; //Z-far

//user variables
int samples = 32; //ao sample count

float radius = 10.0; //ao radius
float aoclamp = 0.5; //depth clamp - reduces haloing at screen edges
bool noise = true; //use noise instead of pattern for sample dithering
float noiseamount = 0.0002; //dithering amount

float diffarea = 0.4; //self-shadowing reduction
float gdisplace = 0.4; //gauss bell center
float aowidth = 8.0; //gauss bell width

bool mist = true; //use mist?
float miststart = 0.0; //mist start
float mistend = 100.0; //mist end

bool onlyAO = false; //use only ambient occlusion pass?
float lumInfluence = 0.7; //how much luminance affects occlusion

//--------------------------------------------------------

vec2 rand(vec2 coord) //generating noise/pattern texture for dithering
{
	float noiseX = ((fract(1.0-coord.s*(width/2.0))*0.25)+(fract(coord.t*(height/2.0))*0.75))*2.0-1.0;
	float noiseY = ((fract(1.0-coord.s*(width/2.0))*0.75)+(fract(coord.t*(height/2.0))*0.25))*2.0-1.0;
	
	if (noise)
	{
		noiseX = clamp(fract(sin(dot(coord ,vec2(12.9898,78.233))) * 43758.5453),0.0,1.0)*2.0-1.0;
		noiseY = clamp(fract(sin(dot(coord ,vec2(12.9898,78.233)*2.0)) * 43758.5453),0.0,1.0)*2.0-1.0;
	}
	return vec2(noiseX,noiseY)*noiseamount;
}

float doMist()
{
	float zdepth = texture2D(bgl_DepthTexture,texCoord.xy).x;
	float depth = -zfar * znear / (zdepth * (zfar - znear) - zfar);
	return clamp((depth-miststart)/mistend,0.0,1.0);
}

float readDepth(in vec2 coord) 
{
	coord.x = clamp(coord.x,0.001,0.999);
    coord.y = clamp(coord.y,0.001,0.999);
	return (2.0 * znear) / (zfar + znear - texture2D(bgl_DepthTexture, coord ).x * (zfar-znear));
}

float compareDepths(in float depth1, in float depth2,inout int far)
{   
	float garea = aowidth; //gauss bell width    
	float diff = (depth1 - depth2)*100.0; //depth difference (0-100)
	//reduce left bell width to avoid self-shadowing 
	if (diff<gdisplace)
	{
	garea = diffarea;
	}else{
	far = 1;
	}
	
	float gauss = pow(2.7182,-2.0*(diff-gdisplace)*(diff-gdisplace)/(garea*garea));
	return gauss;
}   

float calAO(float depth,float dw, float dh)
{   
	//float dd = (1.0-depth)*radius;
	float dd = radius;
	float temp = 0.0;
	float temp2 = 0.0;
	float coordw = bgl_TexCoord.x + dw*dd;
	float coordh = bgl_TexCoord.y + dh*dd;
	float coordw2 = bgl_TexCoord.x - dw*dd;
	float coordh2 = bgl_TexCoord.y - dh*dd;
	
	vec2 coord = vec2(coordw , coordh);
	vec2 coord2 = vec2(coordw2, coordh2);
	
	int far = 0;
	temp = compareDepths(depth, readDepth(coord),far);
	//DEPTH EXTRAPOLATION:
	if (far > 0)
	{
		temp2 = compareDepths(readDepth(coord2),depth,far);
		temp += (1.0-temp)*temp2;
	}
	
	return temp;
} 

void main(void)
{
	vec2 noise = rand(texCoord); 
	float depth = readDepth(texCoord);
	
	float w = (1.0 / width)/clamp(depth,aoclamp,1.0)+(noise.x*(1.0-noise.x));
	float h = (1.0 / height)/clamp(depth,aoclamp,1.0)+(noise.y*(1.0-noise.y));
	
	float pw;
	float ph;
	
	float ao;
	
	float dl = PI*(3.0-sqrt(5.0));
	float dz = 1.0/float(samples);
	float l = 0.0;
	float z = 1.0 - dz/2.0;
	
	for (int i = 0; i <= samples; i ++)
	{     
		float r = sqrt(1.0-z);
		
		pw = cos(l)*r;
		ph = sin(l)*r;
		ao += calAO(depth,pw*w,ph*h);        
		z = z - dz;
		l = l + dl;
	}
	
	ao /= float(samples);
	ao = 1.0-ao;	
	
	if (mist)
	{
	ao = mix(ao, 1.0,doMist());
	}
	
	vec3 color = texture2D(bgl_RenderedTexture,texCoord).rgb;
	
	vec3 lumcoeff = vec3(0.299,0.587,0.114);
	float lum = dot(color.rgb, lumcoeff);
	vec3 luminance = vec3(lum, lum, lum);
	
	vec3 final = vec3(color*mix(vec3(ao),vec3(1.0),luminance*lumInfluence));//mix(color*ao, white, luminance)
	
	if (onlyAO)
	{
	final = vec3(mix(vec3(ao),vec3(1.0),luminance*lumInfluence)); //ambient occlusion only
	}
	
	
	gl_FragColor = vec4(final,1.0); 
	
}
"""


class SSAO(ULFilter):

    def __init__(self, idx: int = None) -> None:
        cam = logic.getCurrentScene().active_camera
        self.settings = {
            'znear': cam.near,
            'zfar': cam.far
        }
        super().__init__(glsl, idx, {
            'znear': self.settings,
            'zfar': self.settings
        })
    
    def update(self):
        super().update()
        cam = logic.getCurrentScene().active_camera
        self.settings = {
            'znear': cam.near,
            'zfar': cam.far
        }




from bge import logic
from bge import render
from bge.types import KX_GameObject as GameObject
from bge.types import KX_PolyProxy
from mathutils import Vector


def ray_data(
    origin: Vector,
    dest: Vector,
    local: bool,
    dist: float
):
    """Get necessary data to calculate the ray.\n
    Not intended for manual use.
    """
    origin = getattr(origin, 'worldPosition', origin)
    dest = getattr(dest, 'worldPosition', dest)
    if local:
        dest = origin + dest
    d = dest - origin
    d.normalize()
    dist = dist if dist else (origin - dest).length
    dest = origin + d * dist
    return d, dist, dest



def raycast(
    caster: GameObject,
    origin: Vector,
    dest: Vector,
    distance: float = 0,
    prop: str = '',
    material: str = '',
    exclude: bool = False,
    xray: bool = False,
    local: bool = False,
    visualize: bool = False
) -> tuple[GameObject, Vector, Vector, Vector]:
    """Raycast from any point to any target

    :param `caster`: casting object, this object will be ignored by the ray.
    :param `origin`: origin point; any vector or list.
    :param `dest`: target point; any vector or list.
    :param `distance`: distance the ray will be cast
    (0 means the ray will only be cast to target).
    :param `prop`: look only for objects with this property.
    :param `material`: look only for objects with this material applied.
    :param `exclude`: invert the selection for `prop` and `material`.
    :param `xray`: look for objects behind others.
    :param `local`: add the target vector to the origin.
    :param `visualize`: show the raycast.

    :returns: (`obj`, `point`, `normal`, `direction`)
    """
    if exclude:
        exclude_prop, prop = prop, ''
    origin = getattr(origin, 'worldPosition', Vector(origin)).copy()
    dest = getattr(dest, 'worldPosition', Vector(dest)).copy()
    direction, distance, dest = ray_data(origin, dest, local, distance)
    obj, point, normal = caster.rayCast(
        dest,
        origin,
        distance,
        prop,
        xray=xray
    )
    if (material and point) or (obj and exclude and exclude_prop in obj):
        bo = obj.blenderObject
        leftover_dist = distance - (origin - point).length
        while (
            material in [
                slot.material.name for
                slot in
                bo.material_slots
            ] or exclude_prop in obj.getPropertyNames() if exclude else
            material not in [
                slot.material.name for
                slot in
                bo.material_slots
            ]
        ) and leftover_dist > 0:
            if not xray:
                obj, point, normal = None, None, None
                break
            elif point:
                old_point = point
                obj, point, normal = obj.rayCast(
                    dest,
                    point,
                    leftover_dist,
                    prop,
                    xray=xray
                )
                if not obj:
                    break
                bo = obj.blenderObject
                leftover_dist -= (origin - old_point).length
            else:
                obj, point, normal = None, None, None
                break

    if visualize:
        line_dest: Vector = direction.copy()
        line_dest.x *= distance
        line_dest.y *= distance
        line_dest.z *= distance
        line_dest = line_dest + origin
        if not obj:
            render.drawLine(
                origin,
                line_dest,
                [1, 0, 0, 1]
            )
        else:
            render.drawLine(
                origin,
                point,
                [0, 1, 0, 1]
            )
    return (obj, point, normal, direction)


def raycast_face(
    caster: GameObject,
    origin: Vector,
    dest: Vector,
    distance: float = 0,
    prop: str = '',
    material: str = '',
    exclude: bool = False,
    xray: bool = False,
    local: bool = False,
    visualize: bool = False
) -> tuple[GameObject, Vector, KX_PolyProxy, Vector]:
    """Raycast from any point to any target. Returns additional face data.

    :param `caster`: casting object, this object will be ignored by the ray.
    :param `origin`: origin point; any vector or list.
    :param `dest`: target point; any vector or list.
    :param `distance`: distance the ray will be cast
    (0 means the ray will only be cast to target).
    :param `prop`: look only for objects with this property.
    :param `material`: look only for objects with this material applied.
    :param `exclude`: invert the selection for `prop` and `material`.
    :param `xray`: look for objects behind others.
    :param `local`: add the target vector to the origin.
    :param `visualize`: show the raycast.

    :returns: (`obj`, `point`, `normal`, `direction`, `face`, `uv`)
    """
    if exclude:
        exclude_prop, prop = prop, ''
    direction, distance, dest = ray_data(origin, dest, local, distance)
    origin = getattr(origin, 'worldPosition', origin)
    obj, point, normal, face, uv = caster.rayCast(
        dest,
        origin,
        distance,
        prop,
        xray=xray,
        poly=2
    )
    if (material and point) or (obj and exclude):
        bo = obj.blenderObject
        leftover_dist = distance - (origin - point).length
        while (
            material in [
                slot.material.name for
                slot in
                bo.material_slots
            ] or exclude_prop in obj.getPropertyNames() if exclude else
            material not in [
                slot.material.name for
                slot in
                bo.material_slots
            ]
        ) and leftover_dist > 0:
            if not xray:
                obj, point, normal = None, None, None
                break
            elif point:
                old_point = point
                obj, point, normal, face, uv = obj.rayCast(
                    dest,
                    point,
                    leftover_dist,
                    prop,
                    xray=xray,
                    poly=2
                )
                if not obj:
                    break
                bo = obj.blenderObject
                leftover_dist -= (origin - old_point).length
            else:
                obj, point, normal = None, None, None
                break

    if visualize:
        line_dest: Vector = direction.copy()
        line_dest.x *= distance
        line_dest.y *= distance
        line_dest.z *= distance
        line_dest = line_dest + origin
        if not obj:
            render.drawLine(
                origin,
                line_dest,
                [1, 0, 0, 1]
            )
        else:
            render.drawLine(
                origin,
                point,
                [0, 1, 0, 1]
            )
    return (obj, point, normal, direction, face, uv)


def raycast_projectile(
    caster: GameObject,
    origin: Vector,
    aim: Vector,
    power: float,
    distance: float = 100,
    resolution: float = .05,
    prop: str = '',
    xray: bool = False,
    local: bool = False,
    visualize: bool = False
) -> tuple[GameObject, Vector, Vector, list]:
    """Raycast along the predicted parabola of a projectile.

    :param `caster`: casting object, this object will be ignored by the ray.
    :param `origin`: origin point; any vector or list.
    :param `aim`: target point; the parabola will start towards this point.
    :param `power`: "speed" of the projectile; a higher values mean further throws
    :param `distance`: total distance the ray will be cast
    :param `resolution`: detail quality of the parabola; higher values mean less detail
    :param `prop`: look only for objects with this property.
    :param `xray`: look for objects behind others.
    :param `local`: add the target vector to the origin.
    :param `visualize`: show the raycast.

    :returns: (`obj`, `point`, `normal`, `points`)
    """
    def calc_projectile(t, vel, pos):
        half: float = logic.getCurrentScene().gravity.z * (.5 * t * t)
        vel = vel * t
        return Vector((0, 0, half)) + vel + pos

    aim.normalize()
    aim *= power
    origin = getattr(origin, 'worldPosition', origin)
    if local:
        origin = origin + caster.worldPosition

    points: list = []
    color: list = [1, 0, 0]
    idx = 0
    total_dist: float = 0

    while total_dist < distance:
        target = (calc_projectile(idx, aim, origin))
        start = origin if not points else points[-1]
        obj, point, normal = caster.rayCast(
            start,
            target,
            prop=prop,
            xray=xray
        )
        total_dist += (target-start).length
        if not obj:
            points.append(target)
        else:
            points.append(point)
            color = [0, 1, 0]
            break
        idx += resolution
    if visualize:
        for i, p in enumerate(points):
            if i < len(points) - 1:
                render.drawLine(p, points[i+1], color)
    return (obj, point, normal, points)


def raycast_camera(
    distance: float = 0,
    prop: str = '',
    xray: bool = False,
    aim: Vector = Vector((.5, .5)),
    visualize: bool = False
):
    """Raycast from any point to any target. Returns additional face data.

    :param `distance`: distance the ray will be cast
    :param `prop`: look only for objects with this property.
    :param `xray`: look for objects behind others.
    :param `aim`: X and Y coordinates of the screen from 0-1

    :returns: (`obj`, `point`, `normal`)
    """
    # assume screen coordinates
    camera = logic.getCurrentScene().active_camera
    if isinstance(aim, Vector) and len(aim) == 2:
        vec = 10 * camera.getScreenVect(aim[0], aim[1])
        ray_target = camera.worldPosition - vec
        aim = ray_target
    if prop:
        obj, point, normal = camera.rayCast(
            aim,
            None,
            distance,
            prop,
            xray=xray
        )
    else:
        obj, point, normal = camera.rayCast(aim, None, distance)
    if visualize:
        if not obj:
            render.drawLine(
                camera.worldPosition,
                aim,
                [1, 0, 0, 1]
            )
        else:
            render.drawLine(
                camera.worldPosition,
                point,
                [0, 1, 0, 1]
            )
    return (obj, point, normal)


def raycast_mouse(
    distance: float = 100,
    prop: str = '',
    material: str = '',
    exclude: bool = False,
    xray: bool = False
):
    """Raycast from the active camera to world cursor coordinates.

    :param `distance`: distance the ray will be cast
    :param `prop`: look only for objects with this property.
    :param `material`: look only for objects with this material applied.
    :param `exclude`: invert the selection for `prop` and `material`.
    :param `xray`: look for objects behind others.
    :param `aim`: X and Y coordinates of the screen from 0-1

    :returns: (`obj`, `point`, `normal`, `direction`)
    """
    camera = logic.getCurrentScene().active_camera
    mpos = logic.mouse.position
    vec = 10 * camera.getScreenVect(*mpos)
    ray_target = camera.worldPosition - vec
    return raycast(
        camera,
        camera.worldPosition,
        ray_target,
        distance,
        prop,
        material,
        exclude,
        xray
    )



from uplogic.nodes import ULActionNode
from uplogic.data import GlobalDB
from uplogic.utils import not_met


class ULStopAllSounds(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        aud_sys = GlobalDB.retrieve('.uplogic_audio').get('nl_audio_system')
        if not aud_sys:
            return
        self._set_ready()
        aud_sys.device.stopAll()
        self._set_value(True)



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.input import ULControllerVR


class ULGetVRControllerValues(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self._index = None
        self.index = None
        self.controller = None
        self.POS = ULOutSocket(self, self.get_pos)
        self.ORI = ULOutSocket(self, self.get_ori)
        self.APOS = ULOutSocket(self, self.get_apos)
        self.AORI = ULOutSocket(self, self.get_aori)
        self.STICK = ULOutSocket(self, self.get_stick)
        self.TRIGGER = ULOutSocket(self, self.get_trigger)
    
    @property
    def index(self):
        return self._index
    
    @index.setter
    def index(self, val):
        self._index = val
        self.controller = ULControllerVR(self.index)

    def get_pos(self):
        return self.controller.position
    
    def get_ori(self):
        return self.controller.orientation.to_euler()
    
    def get_apos(self):
        return self.controller.position_aim
    
    def get_aori(self):
        return self.controller.orientation_aim.to_euler()
    
    def get_stick(self):
        return self.controller.thumbstick
    
    def get_trigger(self):
        return self.controller.trigger

    def evaluate(self):
        self._set_ready()



from bge import render
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode


class ULGetVSync(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.OUT = ULOutSocket(self.get_vsync)

    def get_vsync(self):
        return render.getVsync()

    def evaluate(self):
        self._set_ready()



from .shader import ULFilter
from mathutils import Vector


glsl = """
uniform sampler2D bgl_RenderedTexture;
in vec4 bgl_TexCoord;
vec2 texcoord = bgl_TexCoord.xy;
uniform vec3 color;


vec4 gradient(vec4 coo)
{
	vec4 stripes = coo;
	stripes.r *= color.x;
	stripes.g *= color.y;
	stripes.b *= color.z;
	stripes.a = 1.0;
	return stripes;
}

void main (void) 
{ 		
	vec4 value = texture(bgl_RenderedTexture, texcoord);
		

// 	gl_FragColor = gradient(vec4(clamp(gl_TexCoord[3].s,0.0,1.0)));
	gl_FragColor.rgb = gradient(value).rgb;
	gl_FragColor.a = 1.0;	
}
"""

class Levels(ULFilter):

    def __init__(self, color=(1., 1., 1.), idx: int = None) -> None:
        self.settings = {'color': Vector(color)}
        super().__init__(glsl, idx, {'color': self.settings})



from .shader import ULFilter
from mathutils import Vector


glsl = """
uniform sampler2D bgl_RenderedTexture;
uniform sampler2D bgl_DepthTexture;
in vec4 bgl_TexCoord;

uniform float power;
uniform vec3 color;
uniform float start;
uniform float end;

out vec4 fragColor;

vec4 lerp(vec4 a, vec4 b, float fac){
    return (fac * b) + ((1 - fac) * a);
}

float linearize_depth(vec2 uv){
    float n = start;
    float f = end;
    float z = texture(bgl_DepthTexture, uv).x;
    return (2.0 * n) / (f + n - z * (f - n));
}

void main(){
    float x = bgl_TexCoord[0];
    float y = bgl_TexCoord[1];
    vec2 coords = vec2(x, y);
    vec4 col = texture(bgl_RenderedTexture, coords);
    float d = linearize_depth(coords);
    vec4 black = vec4(0, 0, 0, 1);
    vec4 shade = vec4(color.x, color.y, color.z, 1);
    vec4 depth = vec4(d, d, d, 1);
    depth = depth * shade;
    col = lerp(col, depth, d * power);
    fragColor = col;
}
"""


class Mist(ULFilter):

    def __init__(self, start=.1, end=50.0, color=(0.5, 0.7, 0.9), power=1.0, idx: int = None) -> None:
        self.settings = {
            'start': float(start),
            'end': float(end),
            'color': Vector(color),
            'power': float(power)
        }
        super().__init__(glsl, idx, {
            'start': self.settings,
            'end': self.settings,
            'color': self.settings,
            'power': self.settings
        })

    @property
    def start(self):
        return self.settings['start']
    
    @start.setter
    def start(self, val):
        self.settings['start'] = float(val)

    @property
    def end(self):
        return self.settings['end']
    
    @end.setter
    def end(self, val):
        self.settings['end'] = float(val)
    
    @property
    def power(self):
        return self.settings['power']
    
    @power.setter
    def power(self, val):
        self.settings['power'] = val

    @property
    def color(self):
        return self.settings['color']
    
    @color.setter
    def color(self, val):
        if not isinstance(val, Vector):
            val = Vector(val)
        self.settings['color'] = val


from bge import logic
from bge.types import KX_GameObject as GameObject
import bpy
from bpy.types import Material
from .errors import LogicControllerNotSupportedError


def controller_brick_status(owner, controller_name):
    cont = owner.controllers[controller_name]
    state = (
        owner
        .blenderObject
        .game
        .controllers[controller_name]
        .type
    )
    if not cont.sensors:
        return False
    elif state == 'LOGIC_AND':
        return False not in [sens.positive for sens in cont.sensors]
    elif state == 'LOGIC_OR':
        return True in [sens.positive for sens in cont.sensors]
    elif state == 'LOGIC_NAND':
        return False in [sens.positive for sens in cont.sensors]
    elif state == 'LOGIC_NOR':
        return True not in [sens.positive for sens in cont.sensors]
    elif state == 'LOGIC_XOR':
        return [
            sens.positive
            for sens in
            cont.sensors
        ].count(True) % 2 != 0
    elif state == 'LOGIC_XNOR':
        check = cont.sensors[0].positive
        return False not in [
            sens.positive == check
            for sens in
            cont.sensors
        ]
    else:
        raise LogicControllerNotSupportedError


def create_curve(
    name: str,
    bevel_depth: float = 0.0,
    dimensions: int = 3,
    material: str or Material = None,
    collection: str = None
) -> GameObject:
    """Create a `KX_GameObject` containing a `bpy.types.Curve` object.

    :param `name`: Name of the new `KX_GameObject`.
    :param `bevel_depth`: Define the "thickness" of the curve. This will add
    geometry along the spline.
    :param `dimensions`: Set the coordinate space in which to calculate the
    curve.
    :param `material`: The material to use for bevel geometry.
    :param `collection`: The collection to which to add the curve. Leave at
    `None` to use scene collection.
    """
    bcurve = bpy.data.curves.new(name, 'CURVE')
    bcurve.bevel_depth = bevel_depth
    bcurve.dimensions = f'{dimensions}D'
    bobj = bpy.data.objects.new(name, bcurve)
    if material:
        if isinstance(material, str):
            bobj.data.materials.append(bpy.data.materials[material])
        elif isinstance(material, Material):
            bobj.data.materials.append(material)
    if collection:
        if isinstance(collection, str):
            collection = bpy.data.collections.get(collection, bpy.context.scene.collection)
    elif collection is None:
        collection = bpy.context.scene.collection
    collection.objects.link(bobj)
    game_obj = logic.getCurrentScene().convertBlenderObject(bobj)
    return game_obj


def set_curve_points(
    curve: GameObject,
    points: list
) -> None:
    """Set the curve points of a `KX_GameObject` containing a `bpy.types.Curve` object.

    :param `curve`: `KX_GameObject`
    :param `points`: A list of points to use for the curve.
    """
    bcurve = curve.blenderObject.data
    for spline in bcurve.splines:
        bcurve.splines.remove(spline)
    spline = bcurve.splines.new('POLY')
    pos = curve.worldPosition

    spline.points.add(len(points)-1)
    for p, new_co in zip(spline.points, points):
        p.co = ([
            new_co[0] - pos.x,
            new_co[1] - pos.y,
            new_co[2] - pos.z
        ] + [1.0])


class ULCurve():
    """Wrapper class for creating and handling curves more easily.

    :param `name`: Name of this curve object.
    :param `bevel_depth`: Define the "thickness" of the curve. This will add
    geometry along the spline.
    :param `dimensions`: Set the coordinate space in which to calculate the
    curve.
    :param `material`: The material to use for bevel geometry.
    :param `collection`: The collection to which to add the curve. Leave at
    `None` to use scene collection.
    """

    def __init__(
        self,
        name: str,
        bevel_depth: float = 0.0,
        dimensions: int = 3,
        material: str or Material =None,
        collection: str = None
    ) -> None:
        self.object = create_curve(
            name,
            bevel_depth,
            dimensions,
            material,
            collection
        )

    @property
    def name(self):
        """Name of the game object (Read-Only)."""
        return self.object.name

    @name.setter
    def name(self, val):
        print('ULCurve.name is read-only!')

    @property
    def points(self):
        """Points of the curve. These points use global coordinates."""
        splines = self.object.blenderObject.data.splines
        return splines[0].points if len(splines) > 0 else []

    @points.setter
    def points(self, val):
        if val != self.points:
            set_curve_points(self.object, val)

    @property
    def bevel_depth(self):
        """Thickness of the curve geometry."""
        return self.object.blenderObject.data.bevel_depth

    @bevel_depth.setter
    def bevel_depth(self, val):
        self.object.blenderObject.data.bevel_depth = val



from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
import math


class ULFormula(ULParameterNode):

    @classmethod
    def signum(cls, a):
        return (a > 0) - (a < 0)

    @classmethod
    def curt(cls, a):
        if a > 0:
            return a**(1./3.)
        else:
            return -(-a)**(1./3.)

    def __init__(self):
        ULParameterNode.__init__(self)
        self.a = None
        self.b = None
        self.formula = ""
        self._formula_globals = globals()
        self._formula_locals = {
            "exp": math.exp,
            "pow": math.pow,
            "log": math.log,
            "log10": math.log10,
            "acos": math.acos,
            "asin": math.asin,
            "atan": math.atan,
            "atan2": math.atan2,
            "cos": math.cos,
            "hypot": math.hypot,
            "sin": math.sin,
            "tan": math.tan,
            "degrees": math.degrees,
            "radians": math.radians,
            "acosh": math.acosh,
            "asinh": math.asinh,
            "atanh": math.atanh,
            "cosh": math.cosh,
            "sinh": math.sinh,
            "tanh": math.tanh,
            "pi": math.pi,
            "e": math.e,
            "ceil": math.ceil,
            "sign": ULFormula.signum,
            "abs": math.fabs,
            "floor": math.floor,
            "mod": math.fmod,
            "sqrt": math.sqrt,
            "curt": ULFormula.curt,
            "str": str,
            "int": int,
            "float": float
        }
        self.OUT = ULOutSocket(self, self.get_out)

    def get_out(self):
        socket = self.get_output('out')
        if socket is None:
            a = self.get_input(self.a)
            b = self.get_input(self.b)
            formula_locals = self._formula_locals
            formula_locals["a"] = a
            formula_locals["b"] = b
            out = eval(self.formula, self._formula_globals, formula_locals)
            return self.set_output('out', out)
        return socket

    def evaluate(self):
        self._set_ready()



from .shader import ULFilter

glsl = """
uniform sampler2D bgl_RenderedTexture;
in vec4 bgl_TexCoord;

uniform float power;

void main()
{
   vec2 texcoord = bgl_TexCoord.xy;
   vec3 sum = vec3(0.0);
   vec3 distance = vec3(1.0-(power*0.01), 1.0-(power*0.02), 1.0-(power*0.03));

   sum.r = vec3(texture2D(bgl_RenderedTexture, (texcoord -vec2(0.5,0.5)) * distance[0] + vec2(0.5,0.5))).r;
   sum.g = vec3(texture2D(bgl_RenderedTexture, (texcoord -vec2(0.5,0.5)) * distance[1] + vec2(0.5,0.5))).g;
   sum.b = vec3(texture2D(bgl_RenderedTexture, (texcoord -vec2(0.5,0.5)) * distance[2] + vec2(0.5,0.5))).b;

   gl_FragColor = vec4(sum, 1.0);
}
"""


class ChromaticAberration(ULFilter):

    def __init__(self, power: float = 2.0, idx: int = None) -> None:
        self.settings = {'power': float(power)}
        super().__init__(glsl, idx, {'power': self.settings})
    
    @property
    def power(self):
        return self.settings['power']
    
    @power.setter
    def power(self, val):
        self.settings['power'] = val


from uplogic.nodes import ULConditionNode
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULOnce(ULConditionNode):

    def __init__(self):
        ULConditionNode.__init__(self)
        self.input_condition = None
        self.repeat = None
        self.reset_time = None
        self._consumed = False
        self.time = 0.0

    def evaluate(self):
        condition = self.get_input(self.input_condition)
        repeat = self.get_input(self.repeat)
        reset_time = self.get_input(self.reset_time)
        if is_waiting(repeat, reset_time):
            self._set_value(False)
            return
        network = self.network
        tl = network.timeline

        self._set_ready()
        cond_f = not_met(condition)
        self.time = tl
        if not cond_f and self._consumed is False:
            self._consumed = True
            self._set_value(True)
            return
        if cond_f and repeat and self._consumed:
            self._consumed = False
        self._set_value(False)



from uplogic.nodes import ULActionNode
from uplogic.utils import is_waiting
from uplogic.utils import not_met


class ULSetPyInstanceAttr(ULActionNode):
    def __init__(self):
        ULActionNode.__init__(self)
        self.condition = None
        self.instance = None
        self.attr = None
        self.value = None

    def evaluate(self):
        condition = self.get_input(self.condition)
        if not_met(condition):
            return
        instance = self.get_input(self.instance)
        attr = self.get_input(self.attr)
        value = self.get_input(self.value)
        if is_waiting(instance, attr, value):
            return
        self._set_ready()
        setattr(instance, attr, value)



from uplogic.nodes import ULConditionNode
from uplogic.utils import STATUS_WAITING, is_waiting


class ULOr(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.ca = False
        self.cb = False

    def evaluate(self):
        ca = self.get_input(self.ca)
        cb = self.get_input(self.cb)
        self._set_ready()
        if is_waiting(ca):
            ca = False
        if is_waiting(cb):
            cb = False
        self._set_value(ca or cb)


class ULOrList(ULConditionNode):
    def __init__(self):
        ULConditionNode.__init__(self)
        self.ca = False
        self.cb = False
        self.cc = False
        self.cd = False
        self.ce = False
        self.cf = False

    def evaluate(self):
        ca = self.get_input(self.ca)
        cb = self.get_input(self.cb)
        cc = self.get_input(self.cc)
        cd = self.get_input(self.cd)
        ce = self.get_input(self.ce)
        cf = self.get_input(self.cf)
        self._set_ready()
        conds = [ca, cb, cc, cd, ce, cf]
        self._set_value(True in conds)



from uplogic.animation import ULActionSystem
from uplogic.data import GlobalDB
from uplogic.nodes import ULOutSocket
from uplogic.nodes import ULParameterNode
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_invalid
from uplogic.utils import is_waiting


class ULActionStatus(ULParameterNode):
    def __init__(self):
        ULParameterNode.__init__(self)
        self.game_object = None
        self.action_layer = None
        self._action_name = None
        self._action_frame = None
        self.action = None
        self.act_system = self.get_act_system()
        self.ACTION_NAME = ULOutSocket(self, self.get_action_name)
        self.ACTION_FRAME = ULOutSocket(self, self.get_action_frame)
        
    def get_act_system(self):
        act_systems = GlobalDB.retrieve('uplogic.animation')
        if act_systems.check('default'):
            return act_systems.get('default')
        else:
            return ULActionSystem('default')

    def get_action_name(self):
        if self.action:
            return self.action.name
        return STATUS_WAITING

    def get_action_frame(self):
        if self.action:
            return self.action.frame
        return STATUS_WAITING

    def evaluate(self):
        game_object = self.get_input(self.game_object)
        action_layer = self.get_input(self.action_layer)
        if is_waiting(game_object, action_layer):
            return
        self._set_ready()
        self.action = self.act_system.get_layer(game_object, action_layer)
        if self.action is None:
            self._set_value(False)
            return
        self._set_value(self.action.is_playing)



from bge.types import KX_GameObject as GameObject
from bge import logic
import bpy


def make_unique_light(lamp_ge: GameObject) -> GameObject:
    '''TODO: Documentation
    '''
    lamp = lamp_ge.blenderObject
    lamp.data = lamp.data.copy()
    return lamp_ge


class ULLight():

    def __init__(
        self,
        name: str = '',
        type: str = 'POINT',
        lamp: GameObject = None
    ) -> None:
        if lamp:
            self.light = make_unique_light(lamp)
            return
        game_scene = logic.getCurrentScene()
        scene = bpy.data.scenes[game_scene.name]
        light = bpy.data.lights.new(name, type)
        light = bpy.data.objects.new(name, light)
        scene.collection.objects.link(light)
        self.light = game_scene.convertBlenderObject(light)
        self.energy = 10

    @property
    def energy(self):
        return self.light.blenderObject.data.energy

    @energy.setter
    def energy(self, val):
        self.light.blenderObject.data.energy = val

    @property
    def color(self):
        return self.light.blenderObject.data.color

    @color.setter
    def color(self, val):
        self.light.blenderObject.data.color = val

    @property
    def parent(self):
        return self.light.parent

    @parent.setter
    def parent(self, val):
        self.light.setParent(val)

    @property
    def worldPosition(self):
        return self.light.worldPosition

    @worldPosition.setter
    def worldPosition(self, val):
        self.light.worldPosition = val

    @property
    def localPosition(self):
        return self.light.localPosition

    @localPosition.setter
    def localPosition(self, val):
        self.light.localPosition = val

    @property
    def worldOrientation(self):
        return self.light.worldOrientation

    @worldOrientation.setter
    def worldOrientation(self, val):
        self.light.worldOrientation = val

    @property
    def localOrientation(self):
        return self.light.localOrientation

    @localOrientation.setter
    def localOrientation(self, val):
        self.light.localOrientation = val

    @property
    def worldScale(self):
        return self.light.worldScale

    @worldScale.setter
    def worldScale(self, val):
        self.light.worldScale = val

    @property
    def localScale(self):
        return self.light.localScale

    @localScale.setter
    def localScale(self, val):
        self.light.localScale = val

    @property
    def worldLinearVelocity(self):
        return self.light.worldLinearVelocity

    @worldLinearVelocity.setter
    def worldLinearVelocity(self, val):
        self.light.worldLinearVelocity = val

    @property
    def localLinearVelocity(self):
        return self.light.localLinearVelocity

    @localLinearVelocity.setter
    def localLinearVelocity(self, val):
        self.light.localLinearVelocity = val

    @property
    def worldAngularVelocity(self):
        return self.light.worldAngularVelocity

    @worldAngularVelocity.setter
    def worldAngularVelocity(self, val):
        self.light.worldAngularVelocity = val

    @property
    def localAngularVelocity(self):
        return self.light.localAngularVelocity

    @localAngularVelocity.setter
    def localAngularVelocity(self, val):
        self.light.localAngularVelocity = val

    @property
    def worldTransform(self):
        return self.light.worldTransform

    @worldTransform.setter
    def worldTransform(self, val):
        self.light.worldTransform = val


from mathutils import Vector
from uplogic.nodes import ULParameterNode
from uplogic.nodes import ULOutSocket
from uplogic.utils import STATUS_WAITING
from uplogic.utils import is_waiting


class ULLimitRange(ULParameterNode):

    def __init__(self):
        ULParameterNode.__init__(self)
        self.value = None
        self.threshold = Vector((0, 0))
        self.operator = None
        self.last_val = 0
        self.OUT = ULOutSocket(self, self.get_done)

    def get_done(self):
        socket = self.get_output('done')
        if socket is None:
            v = self.get_input(self.value)
            t = self.get_input(self.threshold)
            if is_waiting(v, t):
                return STATUS_WAITING
            self.calc_threshold(self.operator, v, t)
            if (v is None) or (t is None):
                return STATUS_WAITING
            else:
                return self.set_output('done', self.last_val)
        return socket

    def calc_threshold(self, op, v, t):
        last = self.last_val
        if op == 'OUTSIDE':
            if (v < t.x or v > t.y):
                self.last_val = v
            else:
                self.last_val = t.x if last <= t.x else t.y
        if op == 'INSIDE':
            if (t.x < v < t.y):
                self.last_val = v
            else:
                self.last_val = t.x if v <= t.x else t.y

    def evaluate(self):
        self._set_ready()



from bge import logic
import bpy


def set_scene(scene: str or bpy.types.Scene) -> None:
    """Set a scene.
    """
    if isinstance(scene, str):
        scene = bpy.data.scenes.get(scene)
    if not scene:
        return
    obj = bpy.data.objects.new('Empty', None)
    bpy.context.scene.collection.objects.link(obj)
    bpy.context.view_layer.objects.active = obj
    bpy.ops.logic.sensor_add(
        type='ALWAYS',
        name='Always',
        object=obj.name
    )
    bpy.ops.logic.controller_add(
        type='LOGIC_AND',
        name='And',
        object=obj.name
    )
    bpy.ops.logic.actuator_add(
        type='SCENE',
        name='Set Scene',
        object=obj.name
    )
    always_brick = obj.game.sensors['Always']
    and_brick = obj.game.controllers['And']
    scene_brick = obj.game.actuators['Set Scene']
    scene_brick.mode = 'SET'
    scene_brick.scene = scene

    always_brick.link(and_brick)
    and_brick.link(actuator=scene_brick)

    logic.getCurrentScene().convertBlenderObject(obj)
